/*
 * linux/drivers/s390/cmf.c ($Revision: 1.5.6.2 $)
 *
 * Linux on zSeries Channel Measurement Facility support
 *
 * Copyright 2000,2003 IBM Corporation
 *
 * Author: Arnd Bergmann <arndb@de.ibm.com>
 *
 * original idea from Natarajan Krishnaswami <nkrishna@us.ibm.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/slab.h>

#include <asm/cmb.h>
#include <asm/div64.h>
#include <asm/irq.h>

/**
 * enum cmb_format - types of supported measurement block formats
 *
 * @CMF_BASIC:      traditional channel measurement blocks supported
 * 		    by all machines that we run on
 * @CMF_EXTENDED:   improved format that was introduced with the z990
 * 		    machine
 * @CMF_AUTODETECT: default: use extended format when running on a z990
 *                  or later machine, otherwise fall back to basic format
 **/
enum cmb_format {
	CMF_BASIC,
	CMF_EXTENDED,
	CMF_AUTODETECT = -1,
};
/**
 * format - actual format for all measurement blocks
 * 
 * The format module parameter can be set to a value of 0 (zero)
 * or 1, indicating basic or extended format as described for
 * enum cmb_format.
 */
static int format = CMF_AUTODETECT;
MODULE_PARM(format, "i");

/**
 * struct cmb_operations - functions to use depending on cmb_format
 *
 * all these functions operate on a struct cmf_device. There is only
 * one instance of struct cmb_operations because all cmf_device 
 * objects are guaranteed to be of the same type.
 *
 * @alloc:	allocate memory for a channel measurement block,
 *		either with the help of a special pool or with kmalloc
 * @free:	free memory allocated with @alloc
 * @set:	enable or disable measurement
 * @readall:	read a measurement block in a common format
 * @reset:	clear the data in the associated measurement block and
 *		reset its time stamp
 */
struct cmb_operations {
	int (*alloc)  (struct cmf_device*);
	void(*free)   (struct cmf_device*);
	int (*set)    (struct cmf_device*, u32);
	int (*readall)(struct cmf_device*, struct cmbdata *);
	void (*reset) (struct cmf_device*);
};
static struct cmb_operations *cmbops;

/* our user interface is designed in terms of nanoseconds,
 * while the hardware measures total times in its own
 * unit.*/
static inline u64 time_to_nsec(u32 value)
{
	return ((u64)value) * 128000ull;
}

/*
 * Users are usually interested in average times,
 * not accumulated time.
 * This also helps us with atomicity problems
 * when reading sinlge values.
 */
static inline u64 time_to_avg_nsec(u32 value, u32 count)
{
	u64 ret;

	/* no samples yet, avoid division by 0 */
	if (count == 0)
		return 0;

	/* value comes in units of 128 sec */
	ret = time_to_nsec(value);
	do_div(ret, count);

	return ret;
}

/* activate or deactivate the channel monitor. When area is NULL,
 * the monitor is deactivated. The channel monitor needs to
 * be active in order to measure subchannels, which also need
 * to be enabled. */
static inline void
cmf_activate(void *area, unsigned int onoff)
{
	register void * __gpr2 asm("2");
	register long __gpr1 asm("1");

	__gpr2 = area;
	__gpr1 = onoff ? 2 : 0;
	/* activate channel measurement */
	asm("schm" : : "d" (__gpr2), "d" (__gpr1) );
}

static int
set_schib(int sch, u32 mme, int mbfc, unsigned long address)
{
	int ret;
	int retry;
	schib_t *schib;
	unsigned long mba, mbi;

	/* address can be either a block address or a block index */
	mba = mbi = 0;
	(mbfc ? mba : mbi) = address;

	/* msch can silently fail, so do it again if necessary */
	for (retry = 0; retry < 3; retry++) {
		/* prepare schib */
		schib = &ioinfo[sch]->schib;
		stsch(sch, schib);
		schib->pmcw.mme  = mme;
		schib->pmcw.mbfc = mbfc;
		schib->pmcw.mbi  = mbi;
		if (mbfc)
			schib->mba	 = mba;

		/* try to submit it */
		switch(ret = msch_err(sch, schib)) {
			case 0:
				break;
			case 1:
			case 2: /* in I/O or status pending */
				ret = -EBUSY;
				break;
			case 3: /* subchannel is no longer valid */
				ret = -ENODEV;
				break;
			default: /* msch caught an exception */
				ret = -EINVAL;
				break;
		}
		stsch(sch, schib); /* restore the schib */
		
		if (ret)
			break;
		
		/* check if it worked */
		if (schib->pmcw.mme  == mme &&
		    schib->pmcw.mbfc == mbfc &&
		    schib->pmcw.mbi == mbi &&
		    (!mbfc || schib->mba == mba)) 
			return 0;

		ret = -EINVAL;
	}

	return ret;
}

/**
 * struct cmb_area - container for global cmb data
 *
 * @mem:	pointer to CMBs (only in basic measurement mode)
 * @list:	contains a linked list of all subchannels
 * @lock:	protect concurrent access to @mem and @list
 */
struct cmb_area {
	struct cmb *mem;
	struct list_head list;
	spinlock_t lock;
};
static struct cmb_area cmb_area = {
	.lock = SPIN_LOCK_UNLOCKED,
	.list = LIST_HEAD_INIT(cmb_area.list),
};


/* ****** old style CMB handling ********/

/** int maxchannels
 *
 * Basic channel measurement blocks are allocated in one contiguous
 * block of memory, which can not be moved as long as any channel
 * is active. Therefore, a maximum number of subchannels needs to
 * be defined somewhere. This is a module parameter, defaulting to
 * a resonable value of 1024, or 32 kb of memory.
 * Current kernels don't allow kmalloc with more than 128kb, so the
 * maximum is 4096
 */
static int maxchannels = 1024;
MODULE_PARM(maxchannels,"i");

/**
 * struct cmb - basic channel measurement block
 *
 * cmb as used by the hardware the fields are described in z/Architecture
 * Principles of Operation, chapter 17.
 * The area to be a contiguous array and may not be reallocated or freed.
 * Only one cmb area can be present in the system.
 */
struct cmb {
	u16 ssch_rsch_count;
	u16 sample_count;
	u32 device_connect_time;
	u32 function_pending_time;
	u32 device_disconnect_time;
	u32 control_unit_queuing_time;
	u32 device_active_only_time;
	u32 reserved[2];
};

/* insert a single device into the cmb_area list
 * called with cmb_area.lock held from alloc_cmb
 */
static inline int
alloc_cmb_single (struct cmf_device *cdev)
{
	struct cmb *cmb;
	struct cmf_device *node;
	int ret;

	spin_lock_irq(cdev->ccwlock);
	if (!list_empty(&cdev->cmb_list)) {
		ret = -EBUSY;
		goto out;
	}

	/* find first unused cmb in cmb_area.mem.
	 * this is a little tricky: cmb_area.list
	 * remains sorted by ->cmb pointers */
	cmb = cmb_area.mem;
	list_for_each_entry(node, &cmb_area.list, cmb_list) {
		if ((struct cmb*)node->cmb > cmb)
			break;
		cmb++;
	}
	if (cmb - cmb_area.mem >= maxchannels) {
		ret = -ENOMEM;
		goto out;
	}

	/* insert new cmb */
	list_add_tail(&cdev->cmb_list, &node->cmb_list);
	cdev->cmb = cmb;
	ret = 0;
out:
	spin_unlock_irq(cdev->ccwlock);
	return ret;
}

static int
alloc_cmb (struct cmf_device *cdev)
{
	int ret;
	struct cmb *mem;
	ssize_t size;

	spin_lock(&cmb_area.lock);

	if (!cmb_area.mem) {
		/* there is no user yet, so we need a new area */
		size = sizeof(struct cmb) * maxchannels;
		BUG_ON(!list_empty(&cmb_area.list));

		spin_unlock(&cmb_area.lock);
		mem = kmalloc(size, GFP_KERNEL | GFP_DMA);
		spin_lock(&cmb_area.lock);

		if (cmb_area.mem) {
			/* ok, another thread was faster */
			kfree(mem);
		} else if (!mem) {
			/* no luck */
			ret = -ENOMEM;
			goto out;
		} else {
			/* everything ok */
			memset(mem, 0, size);
			cmb_area.mem = mem;
			cmf_activate(cmb_area.mem, 1);
		}
	}

	/* do the actual allocation */
	ret = alloc_cmb_single(cdev);
out:
	spin_unlock(&cmb_area.lock);

	return ret;
}

static void
free_cmb(struct cmf_device *cdev)
{
	spin_lock(&cmb_area.lock);
	spin_lock_irq(cdev->ccwlock);

	if (list_empty(&cdev->cmb_list)) {
		/* already freed */
		goto out;
	}

	cdev->cmb = NULL;
	list_del_init(&cdev->cmb_list);

	if (list_empty(&cmb_area.list)) {
		cmf_activate(NULL, 0);
		kfree(cmb_area.mem);
		cmb_area.mem = NULL;
	}
out:
	spin_unlock_irq(cdev->ccwlock);
	spin_unlock(&cmb_area.lock);
}

static int
set_cmb(struct cmf_device *cdev, u32 mme)
{
	u16 offset;

	if (!cdev->cmb)
		return -EINVAL;

	offset = mme ? (struct cmb *)cdev->cmb - cmb_area.mem : 0;

	return set_schib(cdev->irq, mme, 0, offset);
}

static int
readall_cmb (struct cmf_device *cdev, struct cmbdata *data)
{
	/* yes, we have to put it on the stack
	 * because the cmb must only be accessed
	 * atomically, e.g. with mvc */
	struct cmb cmb;
	unsigned long flags;
	u64 time;

	spin_lock_irqsave(cdev->ccwlock, flags);
	if (!cdev->cmb) {
		spin_unlock_irqrestore(cdev->ccwlock, flags);
		return -ENODEV;
	}

	cmb = *(struct cmb*)cdev->cmb;
	time = get_clock() - cdev->cmb_start_time;
	spin_unlock_irqrestore(cdev->ccwlock, flags);

	*data = (struct cmbdata) {
		/* we only know values before device_busy_time */
		.size = offsetof(struct cmbdata, device_busy_time),

		/* conver to nanoseconds */
		.elapsed_time = (time * 1000) >> 12,

		/* copy data to new structure */
		.ssch_rsch_count		= cmb.ssch_rsch_count,
		.sample_count			= cmb.sample_count,

		/* time fields are converted to nanoseconds while copying */
		.device_connect_time
			= time_to_nsec(cmb.device_connect_time),
		.function_pending_time
			= time_to_nsec(cmb.function_pending_time),
		.device_disconnect_time
			= time_to_nsec(cmb.device_disconnect_time),
		.control_unit_queuing_time
		 	= time_to_nsec(cmb.control_unit_queuing_time),
		.device_active_only_time
			= time_to_nsec(cmb.device_active_only_time),
	};

	return 0;
}

static void
reset_cmb(struct cmf_device *cdev)
{
	struct cmb *cmb;
	spin_lock_irq(cdev->ccwlock);
	cmb = cdev->cmb;
	if (cmb)
		memset (cmb, 0, sizeof (*cmb));
	cdev->cmb_start_time = get_clock();
	spin_unlock_irq(cdev->ccwlock);
}

static struct cmb_operations cmbops_basic = {
	.alloc	= alloc_cmb,
	.free	= free_cmb,
	.set	= set_cmb,
	.readall= readall_cmb,
	.reset	= reset_cmb,
};

/* ******** extended cmb handling ********/

/**
 * struct cmbe - extended channel measurement block
 *
 * cmb as used by the hardware, may be in any 64 bit physical location,
 * the fields are described in z/Architecture Principles of Operation,
 * third edition, chapter 17.
 */
struct cmbe {
	u32 ssch_rsch_count;
	u32 sample_count;
	u32 device_connect_time;
	u32 function_pending_time;
	u32 device_disconnect_time;
	u32 control_unit_queuing_time;
	u32 device_active_only_time;
	u32 device_busy_time;
	u32 initial_command_response_time;
	u32 reserved[7];
};

static int
alloc_cmbe (struct cmf_device *cdev)
{
	struct cmbe *cmbe;
	cmbe = kmalloc (sizeof (*cmbe), GFP_KERNEL /* | GFP_DMA ? */);
	if (!cmbe)
		return -ENOMEM;

	spin_lock_irq(cdev->ccwlock);
	if (cdev->cmb)
		kfree(cmbe);
	else
		cdev->cmb = cmbe;
	spin_unlock_irq(cdev->ccwlock);

	/* activate global measurement if this is the first channel */
	spin_lock(&cmb_area.lock);
	if (list_empty(&cmb_area.list))
		cmf_activate(NULL, 1);
	list_add_tail(&cdev->cmb_list, &cmb_area.list);
	spin_unlock(&cmb_area.lock);

	return 0;
}

static void
free_cmbe (struct cmf_device *cdev)
{
	spin_lock_irq(cdev->ccwlock);
	if (cdev->cmb)
		kfree(cdev->cmb);
	cdev->cmb = NULL;
	spin_unlock_irq(cdev->ccwlock);

	/* deactivate global measurement if this is the last channel */
	spin_lock(&cmb_area.lock);
	list_del_init(&cdev->cmb_list);
	if (list_empty(&cmb_area.list))
		cmf_activate(NULL, 0);
	spin_unlock(&cmb_area.lock);
}

static int
set_cmbe(struct cmf_device *cdev, u32 mme)
{
	unsigned long mba;

	if (!cdev->cmb)
		return -EINVAL;

	mba = mme ? (unsigned long)cdev->cmb : 0;

	return set_schib(cdev->irq, mme, 1, mba);
}

static int
readall_cmbe (struct cmf_device *cdev, struct cmbdata *data)
{
	/* yes, we have to put it on the stack
	 * because the cmb must only be accessed
	 * atomically, e.g. with mvc */
	struct cmbe cmb;
	unsigned long flags;
	u64 time;

	spin_lock_irqsave(cdev->ccwlock, flags);
	if (!cdev->cmb) {
		spin_unlock_irqrestore(cdev->ccwlock, flags);
		return -ENODEV;
	}

	cmb = *(struct cmbe*)cdev->cmb;
	time = get_clock() - cdev->cmb_start_time;
	spin_unlock_irqrestore(cdev->ccwlock, flags);
	
	*data = (struct cmbdata) {
		/* we only know values before device_busy_time */
		.size = offsetof(struct cmbdata, device_busy_time),

		/* conver to nanoseconds */
		.elapsed_time = (time * 1000) >> 12,

		/* copy data to new structure */
		.ssch_rsch_count		= cmb.ssch_rsch_count,
		.sample_count			= cmb.sample_count,

		/* time fields are converted to nanoseconds while copying */
		.device_connect_time
			= time_to_nsec(cmb.device_connect_time),
		.function_pending_time
			= time_to_nsec(cmb.function_pending_time),
		.device_disconnect_time
			= time_to_nsec(cmb.device_disconnect_time),
		.control_unit_queuing_time
		 	= time_to_nsec(cmb.control_unit_queuing_time),
		.device_active_only_time
			= time_to_nsec(cmb.device_active_only_time),
		.device_busy_time
			= time_to_nsec(cmb.device_busy_time),
		.initial_command_response_time
			= time_to_nsec(cmb.initial_command_response_time),
	};

	return 0;
}

static void
reset_cmbe(struct cmf_device *cdev)
{
	struct cmbe *cmb;
	spin_lock_irq(cdev->ccwlock);
	cmb = cdev->cmb;
	if (cmb)
		memset (cmb, 0, sizeof (*cmb));
	cdev->cmb_start_time = get_clock();
	spin_unlock_irq(cdev->ccwlock);
}

static struct cmb_operations cmbops_extended = {
	.alloc	= alloc_cmbe,
	.free	= free_cmbe,
	.set	= set_cmbe,
	.readall= readall_cmbe,
	.reset	= reset_cmbe,
};

/******* external interface to kernel *******/

/**
 * enable_cmf, disable_cmf, set_cmf, cmf_readall, cmf_reset:
 *  simple wrappers around the cmb_operations.
 */
int
enable_cmf(struct cmf_device *cdev)
{
	return cmbops->alloc(cdev);
}

void
disable_cmf(struct cmf_device *cdev)
{
	cmbops->free(cdev);
}

int
set_cmf(struct cmf_device *cdev, u32 mme)
{
	return cmbops->set(cdev, mme);
}

int
cmf_readall(struct cmf_device *cdev, struct cmbdata *data)
{
	return cmbops->readall(cdev, data);
}

void
cmf_reset(struct cmf_device *cdev)
{
	return cmbops->reset(cdev);
}

/* set up maxchannels and format parameter when the module is built-in */
#ifndef MODULE
static int __init
setup_cmf_maxchannels(char *arg)
{
	int arglen = sizeof("cmf_maxchannels=") - 1;
	int c;
	
	c = simple_strtoul(arg + arglen + 1, 0, 0);
	if (c <= 0 || c > 65536) {
		printk(KERN_WARNING "Invalid parameter %s, using "
			"default (%d)\n", arg, maxchannels);
	} else {
		maxchannels = c;
	}

	return 0;
}
__setup("cmf_maxchannels=", setup_cmf_maxchannels);

static int __init
setup_cmf_format(char *arg)
{
	int arglen = sizeof("cmf_format") - 1;
	
	format = simple_strtoul(arg + arglen, 0, 0);
	if (format < CMF_BASIC || format > CMF_EXTENDED) {
		printk(KERN_WARNING "Invalid parameter %s\n", arg);
		format = -1;
	}

	return 0;
}
__setup("cmf_format=", setup_cmf_format);
#endif

static int __init
init_cmf(void)
{
	char *format_string;
	char *detect_string = "parameter";

	/* We cannot really autoprobe this. If the user did not give a parameter,
	   see if we are running on z990 or up, otherwise fall back to basic mode. */

	if (format == CMF_AUTODETECT) {
		if (!MACHINE_NEW_STIDP) {
			format = CMF_BASIC;
		} else {
			format = CMF_EXTENDED;
		}
		detect_string = "autodetected";
	} else {
		detect_string = "parameter";
	}

	switch (format) {
	case CMF_BASIC:
		format_string = "basic";
		cmbops = &cmbops_basic;
		if (maxchannels > 4096 || maxchannels < 1) {
			printk(KERN_ERR "Basic channel measurement facility"
					" can only use 1 to 4096 devices\n"
			       KERN_ERR "when the cmf driver is built"
					" as a loadable module\n");
			return 1;
		}
		break;
	case CMF_EXTENDED:
 		format_string = "extended";
		cmbops = &cmbops_extended;
		break;
	default:
		printk(KERN_ERR "Invalid format %d for channel "
			"measurement facility\n", format);
		return 1;
	}

	printk(KERN_INFO "Channel measurement facility using %s format (%s)\n",
		format_string, detect_string);
	return 0;
}

module_init(init_cmf);


MODULE_AUTHOR("Arnd Bergmann <arndb@de.ibm.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("channel measurement facility base driver\n"
		   "Copyright 2003 IBM Corporation\n");

EXPORT_SYMBOL_GPL(enable_cmf);
EXPORT_SYMBOL_GPL(disable_cmf);
EXPORT_SYMBOL_GPL(set_cmf);
EXPORT_SYMBOL_GPL(cmf_readall);
EXPORT_SYMBOL_GPL(cmf_reset);
